import os
import re
import warnings
from dataclasses import dataclass
from datetime import date
from functools import cache
from glob import iglob
from pathlib import Path
from subprocess import check_output
from typing import List, NamedTuple, Optional

import frontmatter
from jinja2 import Environment, FileSystemLoader, Template, select_autoescape
from markdown import Markdown
from markdown_grid_tables import GridTableExtension
from weasyprint import HTML

from . import extensions


class Article(NamedTuple):
    source: Path
    title: str
    content: str
    meta: dict[str, object]
    has_custom_headline: bool

    @property
    def hash(self):
        return str(check_output(["git", "hash-object", self.source]), "utf-8")

    @property
    def modified_date(self):
        return str(check_output(["git", "log", "-1", "--pretty=%cs", self.source]), "utf-8").strip()


@dataclass
class Document:
    title: str
    template: Template
    layout_dir: Path
    articles: List[Article]
    meta: dict[str, object]

    @staticmethod
    def get_commit():
        if commit_sha_env := os.getenv("CI_COMMIT_SHORT_SHA", None):
            return commit_sha_env

        return str(check_output(["git", "rev-parse", "HEAD"]), "utf-8")[:8] + ("-dirty" if check_output(["git", "status", "-s"]) else "")

    def write_pdf(self, output_dir: Path, output_html: bool = False):
        html = self.template.render(
            date=date.today().isoformat(),
            commit=self.get_commit(),
            articles=self.articles,
            title=self.title,
            meta=self.meta,
        )

        output_filename = self.title.replace(" ", "_")
        if output_html:
            with open(output_dir / f"{output_filename}.html", "w", encoding="utf-8") as html_file:
                html_file.write(html)

        pdf_output_target = output_dir / f"{output_filename}.pdf"
        HTML(string=html, base_url=str(self.layout_dir)).write_pdf(target=pdf_output_target)
        return pdf_output_target


class Printer:
    @staticmethod
    def _ensure_path(path: Path, dir: Optional[bool] = None, create: Optional[bool] = None):
        if not path.is_absolute():
            path = Path(os.path.join(os.getcwd(), path))

        if not path.exists():
            if create and dir:
                path.mkdir(parents=True)

            else:
                raise FileNotFoundError("Path does not exist")

        if dir is True and not path.is_dir():
            raise ValueError(f"{path} is not a directory")

        return path

    def __init__(
        self,
        input: Path,
        output_dir: Path,
        layouts_dir: Path = Path("layouts"),
        bundle: bool = False,
        title: Optional[str] = None,
        layout: Optional[str] = None,
        output_html: bool = False,
        filename_filter: Optional[str] = None,
        meta: Optional[dict[str, object]] = None,
    ):
        self.input = self._ensure_path(input)
        self.output_dir = self._ensure_path(output_dir, dir=True, create=True)
        self.layouts_dir = self._ensure_path(layouts_dir, dir=True)
        self.bundle = bundle
        self.title = title
        self.layout = layout
        self.output_html = output_html
        self.filename_filter = re.compile(filename_filter) if filename_filter else None
        self.meta = meta or {}
        self.jinja_env = Environment(
            autoescape=select_autoescape(),
            loader=FileSystemLoader(searchpath=[self.layouts_dir]),
        )

        if self.bundle:
            if not self.layout or not self.title:
                raise ValueError("A layout and title must be specified when using bundle.")

            if not os.path.isdir(self.input):
                warnings.warn("Option bundle has no effect when using a single file as input")

        elif not self.bundle:
            if self.title:
                raise ValueError("A title cannot be specified when not using bundle.")

    def _load_article(self, source: Path | str):
        if isinstance(source, str):
            source = Path(source)

        with open(source, mode="r", encoding="utf-8") as file:
            article = frontmatter.load(file)

        md = Markdown(
            extensions=[
                extensions.FootnoteExtension(),
                extensions.TableExtension(),
                extensions.ToaExtension(),
                extensions.AbbrExtension(),
                extensions.TocExtension(id_prefix=source.name, toc_depth="2-6"),
                extensions.SubscriptExtension(),
                extensions.TextboxExtension(),
                extensions.CheckboxExtension(),
                GridTableExtension(),
            ],
        )

        content = (
            Environment(
                autoescape=select_autoescape(),
                loader=FileSystemLoader(searchpath=[os.path.dirname(source), self.input, os.getcwd()]),
            )
            .from_string(article.content)
            .render()
        )

        return Article(
            source=source,
            title=source.name.removesuffix(".md").replace("_", " "),
            content=md.convert(content),
            meta=article.metadata,
            has_custom_headline=content.startswith("# "),
        )

    def execute(self):
        self._load_template.cache_clear()
        articles: List[Article] = []
        if self.input.is_dir():
            for article_path in sorted(iglob(os.path.join(self.input, "**/*.md"), recursive=True)):
                if os.path.basename(article_path).startswith("_"):
                    continue

                if self.filename_filter and not re.search(self.filename_filter, Path(article_path).relative_to(self.input).as_posix()):
                    continue

                articles.append(self._load_article(article_path))

        else:
            articles.append(self._load_article(self.input))

        write_options = {"output_dir": self.output_dir, "output_html": self.output_html}

        if self.bundle:
            doc = Document(
                self.title,  # type: ignore  # title cannot be empty when bundle is set
                *self._load_template(self.layout),
                articles=articles,
                meta=self.meta,
            )
            yield doc, doc.write_pdf(**write_options)

        else:
            for article in articles:
                try:
                    doc = Document(
                        article.title,
                        *self._load_template(article.meta.get('layout', self.layout)),
                        articles=[article],
                        meta=self.meta | article.meta,
                    )

                except ValueError as error:
                    raise ValueError(f"Could not create document for {article.source}: {error}") from error

                yield doc, doc.write_pdf(**write_options)

    def _get_layout_dir(self, layout: str):
        if not layout:
            raise ValueError("No layout defined")

        if os.path.isdir(layout_dir := self.layouts_dir / layout):
            return layout_dir

        raise ValueError("Layout \"{layout}\" could not be found")

    @cache
    def _load_template(self, layout):
        layout_dir = self._get_layout_dir(layout)
        with open(layout_dir / "index.html", mode="rb") as file:
            template = self.jinja_env.from_string(str(file.read(), "utf-8"))

        return template, layout_dir



import re
import xml.etree.ElementTree as etree
from typing import Dict

from markdown import Markdown
from markdown.blockparser import BlockParser
from markdown.blockprocessors import BlockProcessor
from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor


class ToaExtension(Extension):
    """Table of Abbreviations Extension for Python-Markdown."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def extendMarkdown(self, md):
        """Insert after AbbrPreprocessor."""
        md.treeprocessors.register(ToaProcessor(md.parser, self), 'toa', 5)


class ToaProcessor(Treeprocessor):
    def __init__(self, md: Markdown, toa: ToaExtension) -> None:
        self.marker = "[TOA]"
        self.toa = toa
        super().__init__(md)

    def iterparent(self, node):
        '''Iterator wrapper to get allowed parent and child all at once.'''

        for child in node:
            if child.tag not in ['pre', 'code']:
                yield node, child
                yield from self.iterparent(child)

    def replace_marker(self, root, elem):
        for p, c in self.iterparent(root):
            text = ''.join(c.itertext()).strip()
            if not text:
                continue

            if c.text and c.text.strip() == self.marker and len(c) == 0:
                for i in range(len(p)):
                    if p[i] == c:
                        p[i] = elem
                        break

    def build_toa_table(self, abbreviations: Dict[str, str]) -> etree.Element:
        table = etree.Element("table", attrib={"class": "toa"})
        tbody = etree.SubElement(table, "tr")
        for title, text in abbreviations.items():
            tr = etree.SubElement(tbody, "tr")
            td = etree.SubElement(tr, "td")
            td.text = title
            td = etree.SubElement(tr, "td")
            td.text = text

        return table

    def run(self, doc: etree.Element) -> None:
        abbreviations = {}
        for el in doc.iter("abbr"):
            abbreviations[el.text] = el.get("title")

        self.replace_marker(doc, self.build_toa_table(abbreviations))



import json
import time
import warnings
from functools import partial
from pathlib import Path
from threading import Timer
from typing import Callable, Optional

import typer
from rich.console import Console
from typing_extensions import Annotated
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from .printer import Printer

console = Console()

warnings.showwarning = lambda message, category, filename, lineno, file, line: console.print("Warning:", message, style="bold yellow")


def debounce(wait):
    def decorator(fn):
        def debounced(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)

            try:
                debounced.t.cancel()

            except AttributeError:
                pass

            debounced.t = Timer(wait, call_it)
            debounced.t.start()

        return debounced

    return decorator


class FileChangeHandler(FileSystemEventHandler):
    def __init__(self, render: Callable[[], None]) -> None:
        self._render = render

    @debounce(1)
    def render(self):
        self._render()
        console.log("Render complete")

    def on_created(self, event: FileSystemEvent):
        console.log("Rerender")
        try:
            self.render()

        except Exception:
            console.print_exception()

    def on_modified(self, event: FileSystemEvent):
        self.on_created(event)


def main(
    input: Annotated[Path, typer.Argument(help="Folder or file used as input")],
    output_dir: Annotated[Path, typer.Argument(help="Folder where resulting files are written to")] = Path("."),
    *,
    bundle: Annotated[bool, typer.Option(help="Bundle all input documents into a single output file")] = False,
    title: Annotated[Optional[str], typer.Option(help="Title of the resulting document. Can only be used in conjunction with bundle.")] = None,
    layouts_dir: Annotated[Path, typer.Option(help="Base folder containing the available layouts")] = Path("./layouts"),
    layout: Annotated[Optional[str], typer.Option(help="Default layout to use")] = None,
    output_html: Annotated[bool, typer.Option(help="Additionally output the raw HTML file which is used to create the pdf")] = False,
    filename_filter: Annotated[
        Optional[str],
        typer.Option(help="Regular expression to filter files in input directory by subpath and/or filename"),
    ] = None,
    meta: Annotated[
        Optional[str],
        typer.Option(help="Metadata for document generation passed to the layout, pass values using a JSON object"),
    ] = None,
    watch: Annotated[bool, typer.Option(help="Watch input directory for changes and re-run the conversion")] = False,
):
    try:
        printer = Printer(
            input=input,
            output_dir=output_dir,
            layouts_dir=layouts_dir,
            bundle=bundle,
            title=title,
            layout=layout,
            output_html=output_html,
            filename_filter=filename_filter,
            meta=json.loads(meta) if meta else None,
        )

    except ValueError as error:
        console.print("Error:", error, style="bold red")
        raise typer.Exit(1)

    def execute():
        try:
            for document, output_path in printer.execute():
                console.log(
                    f"Created PDF",
                    f'"{document.title}"',
                    f"(out of {len(document.articles)})" if len(document.articles) > 1 else f"(from {document.articles[0].source})",
                    "->",
                    output_path,
                )

            console.log("Completed document creation")

        except ValueError as error:
            console.log("Error:", error, style="bold red")

        except Exception:
            console.print_exception()

    if watch:
        observer = Observer()
        add_watch_directory = partial(observer.schedule, FileChangeHandler(execute), recursive=True)
        add_watch_directory(input)
        add_watch_directory(layouts_dir)

        observer.start()

        execute()

        try:
            while True:
                time.sleep(1)

        finally:
            observer.stop()
            observer.join()

            return

    execute()


if __name__ == "__main__":
    typer.run(main)



from markdown.extensions.abbr import AbbrExtension
from markdown.extensions.tables import TableExtension

from .checkbox import CheckboxExtension
from .footnotes import FootnoteExtension
from .subscript import SubscriptExtension
from .textbox import TextboxExtension
from .toa import ToaExtension
from .toc import TocExtension



import re
from xml.etree import ElementTree as etree
from xml.etree.ElementTree import Element

from markdown import Markdown
from markdown.extensions import Extension
from markdown.inlinepatterns import InlineProcessor


class CheckboxInlineProcessor(InlineProcessor):
    def handleMatch(self, m: re.Match[str], data) -> tuple[Element, int, int] | tuple[None, None, None]:
        el = etree.Element("input")
        el.attrib["type"] = "checkbox"
        if m.group(1) == "x":
            el.attrib["checked"] = "checked"

        if m.group(2):
            el.attrib["name"] = m.group(2)

        return el, m.start(0), m.end(0)


class CheckboxExtension(Extension):
    def extendMarkdown(self, md: Markdown) -> None:
        md.inlinePatterns.register(
            CheckboxInlineProcessor(r"\[( |x)?\](\w+)?"),
            "checkbox",
            150,
        )
